@@ -93,21 +93,32 @@ static bundle_t			fdslusb_context[2];
 static per_ctrl_t		fdslusb_params[2];
 static int			started			= FALSE;
 static int			nvers			= 0;
-static struct tq_struct		tq_dpc;			/* for tq_immediate */
-static struct tq_struct		tq_tx_dpc;		/* for tq_private */
-static struct tq_struct		tq_rx_dpc;		/* for tq_private */
 static atomic_t			scheduler_enabled	= ATOMIC_INIT (1);
 static atomic_t			crit_count		= ATOMIC_INIT (0);
 static volatile int		hard_error_issued;
-static unsigned long		qtflags;
-static spinlock_t		qtlock			= SPIN_LOCK_UNLOCKED;
-static atomic_t			qtlockcount		= ATOMIC_INIT (0);
+static atomic_t			tx_flag			= ATOMIC_INIT (0);
+static atomic_t			rx_flag			= ATOMIC_INIT (0);
+static unsigned long		qt_flags;
+static spinlock_t		qt_lock			= SPIN_LOCK_UNLOCKED;
+static spinlock_t		stack_lock		= SPIN_LOCK_UNLOCKED;
 static int			thread_pid		= -1;
 static atomic_t			thread_run_flag;
 static atomic_t			thread_capi_run_flag;
 int				card_config;
 
-static DECLARE_TASK_QUEUE(tq_private);
+static void			tx_task (unsigned long data);
+static void			rx_task (unsigned long data);
+static void			tx_handler (dev_context_p pdc);
+static void			rx_handler (dev_context_p pdc);
+static void			enable_thread ();
+static void 			disable_thread ();
+static void			kick_scheduler (void);
+static int			scheduler (void *);
+void				usb_TxStart (void);
+static void			usb_RxStart (void);
+
+static DECLARE_TASKLET(tx_tasklet, tx_task, 0);
+static DECLARE_TASKLET(rx_tasklet, rx_task, 0);
 static DECLARE_WAIT_QUEUE_HEAD(wait);
 static DECLARE_WAIT_QUEUE_HEAD(capi_wait);
 static DECLARE_MUTEX_LOCKED(hotplug);
@@ -134,34 +145,24 @@ typedef struct {
 
 	struct urb *		urb;
 	char *			buf;
-	int			idx;
-} ubi_t, * ubi_p;
+	int			flag;
+} ubf_t, * ubf_p;
 
-static spinlock_t		ullock			= SPIN_LOCK_UNLOCKED;
-
-static ubi_t			ubi[MAX_URB_COUNT + 1];
-static int			ubi_free		= MAX_URB_COUNT;
-static int			ubi_in			= MAX_URB_COUNT;
-static int			ubi_out			= MAX_URB_COUNT;
-static int			ubi_n			= -1;
-#if !defined (NDEBUG)
-static int			ubi_nmin		= MAX_URB_COUNT;
-#endif
+static ubf_t			UBF[MAX_URB_COUNT];
+static int			free_in;
+static int			free_out;
+static atomic_t			free_n			= ATOMIC_INIT (-1);
+static int			free_ubf[MAX_URB_COUNT];
+static int			todo_in;
+static int			todo_out;
+static atomic_t			todo_n;
+static int			todo_ubf[MAX_URB_COUNT];
 
 #define	URB_SUBMITTED		-1
-#define	URB_UNLINKED		-2
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/
-static void			dpc (void *);
-static void			tx_dpc (void *);
-static void			rx_dpc (void *);
-static void			enable_thread ();
-static void 			disable_thread ();
-static void			kick_scheduler (void);
-static int			scheduler (void *);
-void				usb_TxStart (void);
-static void			usb_RxStart (void);
+#define	URB_COMPLETED		-2
+#define	URB_PROCESSING		-3
+#define	URB_DONE		-4
+#define	URB_UNLINKED		-5
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
@@ -262,8 +263,8 @@ static int nbchans (struct capi_ctr * ctrl) {
 	unsigned char *	prf;
 	int		temp = 2;
     
-	assert (ctrl);
-	assert (ctrl->driverdata);
+	assert (ctrl != NULL);
+	assert (ctrl->driverdata != NULL);
 	pcp = GET_CTRL (ctrl);
 	prf = (unsigned char *) pcp->string[6];
 	if (prf != NULL) {
@@ -492,7 +493,6 @@ static int start (dev_context_p card) {
 	card->count = 0;
 	table_init (&card->appls);
 	queue_init (&card->queue);
-	enable_thread ();
 	(*fdslusb_capi_lib->cm_register_ca_functions) (&ca_funcs);
 	if ((*fdslusb_capi_lib->cm_start) ()) {
 		log ("Starting the stack failed...\n");
@@ -508,6 +508,7 @@ static int start (dev_context_p card) {
 		return FALSE;
 	}
 	leave_critical ();
+	enable_thread ();
 	return (started = TRUE); 
 } /* start */
 
@@ -594,7 +595,7 @@ void start_closing_task (void (* func) (void *), void * data) {
 	spin_lock_irqsave (&close_lock, flags);
 	assert ((close_func1 == NULL) || (close_func2 == NULL));
 	if (close_func1 != NULL) {
-		assert (close_func2 != NULL);
+		assert (close_func2 == NULL);
 		close_func2 = func;
 		close_data2 = data;
 	} else {
@@ -621,44 +622,23 @@ void start_closing_task (void (* func) (void *), void * data) {
 /*---------------------------------------------------------------------------*\
 \*-L-------------------------------------------------------------------------*/
 void lock (void) {
-	unsigned long	local_flags;
-	
-	if (!atomic_inc_and_test (&qtlockcount)) {
-		spin_lock_irqsave (&qtlock, local_flags);
-		qtflags = local_flags;
-	}
+	unsigned long   local_flags;
+
+	info (!spin_is_locked (&qt_lock));
+	spin_lock_irqsave (&qt_lock, local_flags);
+	qt_flags = local_flags;
 } /* lock */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 void unlock (void) {
-	unsigned long	local_flags = qtflags;
+	unsigned long   local_flags = qt_flags;
 
-	if (!atomic_dec_and_test (&qtlockcount)) {
-		spin_unlock_irqrestore (&qtlock, local_flags);
-	}
+	spin_unlock_irqrestore (&qt_lock, local_flags);
 } /* unlock */
 
 /*---------------------------------------------------------------------------*\
 \*-C-------------------------------------------------------------------------*/ 
-static void kick_queue (struct tq_struct * task) {
-
-	if (NULL != task) {
-		queue_task (task, &tq_private);
-	}
-	queue_task (&tq_dpc, &tq_immediate);
-	mark_bh (IMMEDIATE_BH);
-} /* kick_queue */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
-void enter_critical (void) {
-
-	atomic_inc (&crit_count);
-} /* enter_critical */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
 static inline int in_critical (void) {
 
 	return (0 < atomic_read (&crit_count));
@@ -666,18 +646,41 @@ static inline int in_critical (void) {
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/ 
-void leave_critical (void) {
-        static spinlock_t	lock = SPIN_LOCK_UNLOCKED;
-	unsigned long		flags;
-
-	assert (in_critical ());
-	spin_lock_irqsave (&lock, flags);
-	if (1 == atomic_read (&crit_count)) {
-		/* About to leave critical area... */
-		kick_queue (NULL);
+static inline void check (void) {
+	unsigned long	flags;
+	spinlock_t	chk_lock = SPIN_LOCK_UNLOCKED;
+	
+	if (atomic_xchg (&rx_flag, 0)) {
+		spin_lock_irqsave (&chk_lock, flags);
+		rx_handler (fdslusb_capi_ctx);
+		spin_unlock_irqrestore (&chk_lock, flags);
 	}
+	if (atomic_xchg (&tx_flag, 0)) {
+		spin_lock_irqsave (&chk_lock, flags);
+		tx_handler (fdslusb_capi_ctx);
+		spin_unlock_irqrestore (&chk_lock, flags);
+	}
+} /* check */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/ 
+void enter_critical (void) {
+
+	if (!atomic_read (&crit_count) && !in_interrupt ()) {
+		check ();
+	}
+	atomic_inc (&crit_count);
+} /* enter_critical */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/ 
+void leave_critical (void) {
+	
+	assert (in_critical ());
 	atomic_dec (&crit_count);
-	spin_unlock_irqrestore (&lock, flags);
+	if (!atomic_read (&crit_count) && !in_interrupt ()) {
+		check ();
+	}
 } /* leave_critical */
 
 /*---------------------------------------------------------------------------*\
@@ -832,26 +835,6 @@ int __entry fdslusb_add_card (struct capi_driver * drv, capicardparams * args) {
 	UNUSED_ARG (args);
 	init_waitqueue_head (&wait);
 	atomic_set (&resetting_ctrl, 0);
-
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
-	tq_dpc.next = 
-	tq_tx_dpc.next = 
-	tq_rx_dpc.next = NULL;
-#else
-	tq_dpc.list.next = 
-	tq_tx_dpc.list.next = 
-	tq_rx_dpc.list.next = NULL;
-#endif
-	tq_dpc.sync = 
-	tq_tx_dpc.sync =
-	tq_rx_dpc.sync = 0;
-	tq_dpc.data = 
-	tq_tx_dpc.data =
-	tq_rx_dpc.data = fdslusb_capi_ctx;
-	tq_dpc.routine = &dpc;
-	tq_tx_dpc.routine = &tx_dpc;
-	tq_rx_dpc.routine = &rx_dpc;
-						
 	return fdslusb_kcapi_init ();
 } /* fdslusb_add_card */ 
 
@@ -913,7 +896,7 @@ void __entry fdslusb_remove_ctrl (struct capi_ctr * ctrl) {
 	assert (fdslusb_capi_ctrl[1] != NULL);
 	enter_critical ();
 	usb_unlink_urb (fdslusb_capi_ctx->tx_urb);
-	unlink_urbs (fdslusb_capi_ctx);
+	unlink_urbs ();
 	leave_critical ();
 	if (fdslusb_context[0].ctrl != NULL) {
 		(* fdslusb_drv_interface->detach_ctr) (fdslusb_capi_ctrl[0]);
@@ -1009,7 +992,7 @@ void __entry fdslusb_send_msg (struct capi_ctr * ctrl, struct sk_buff * skb) {
 
 	assert (ctrl != NULL);
 	assert (ctrl->driverdata != NULL);
-	assert (skb);
+	assert (skb != NULL);
 	card = GET_CARD (ctrl);
 	byte = skb->data;
 	appl = byte[2] + 256 * byte[3];
@@ -1153,7 +1136,7 @@ struct capi_ctr * card_by_id (unsigned appl_id) {
 void * first_data (int * res) {
 	appl_t * appp;
 
-	assert (res);
+	assert (res != NULL);
 	appp = first_appl (fdslusb_capi_ctx->appls);
 	*res = (appp != NULL) ? 0 : -1;
 	return (appp != NULL) ? appp->data  : NULL;
@@ -1164,7 +1147,7 @@ void * first_data (int * res) {
 void * next_data (int * res) {
 	appl_t * appp;
 
-	assert (res);
+	assert (res != NULL);
 	if (NULL != (appp = get_appl (fdslusb_capi_ctx->appls, *res))) {
 		appp = next_appl (fdslusb_capi_ctx->appls, appp);
 	}
@@ -1181,8 +1164,8 @@ int appl_profile (unsigned appl_id, unsigned * bs, unsigned * bc) {
 	if (NULL == appp) {
 		return 0;
 	}
-	if (bs) *bs = appp->blk_size;
-	if (bc) *bc = appp->blk_count;
+	if (bs != NULL) *bs = appp->blk_size;
+	if (bc != NULL) *bc = appp->blk_count;
 	return 1;
 } /* appl_profile */
 
@@ -1199,11 +1182,11 @@ int msg2stack (unsigned char * msg) {
 	unsigned		temp;
 	int			res = 0;
 
-	assert (msg);
+	assert (msg != NULL);
 	if (!queue_is_empty (fdslusb_capi_ctx->queue)) {
 		skb = queue_peek (fdslusb_capi_ctx->queue);
 		res = 1;
-		assert (skb);
+		assert (skb != NULL);
 		mptr = (unsigned char *) skb->data;
 		mlen = mptr[0] + 256 * mptr[1]; 
 		appl = mptr[2] + 256 * mptr[3];
@@ -1223,11 +1206,11 @@ int msg2stack (unsigned char * msg) {
 			*dptr++ = temp & 0xFF;	temp >>= 8;
 			*dptr++ = temp & 0xFF;	temp >>= 8;
 			*dptr++ = temp & 0xFF;
+			memcpy (msg, mptr, mlen);
 			queue_park (fdslusb_capi_ctx->queue, appl, ncci, hand);
-			memcpy (msg, mptr, mlen);
 		} else {
-			queue_drop (fdslusb_capi_ctx->queue);
 			memcpy (msg, mptr, mlen);
+			queue_drop (fdslusb_capi_ctx->queue);
 		}
 	} 
 	return res;
@@ -1247,7 +1230,7 @@ void msg2capi (unsigned char * msg) {
 	struct sk_buff *	skb;
 	struct capi_ctr *	ctrl;
 
-	assert (msg);
+	assert (msg != NULL);
 	mlen = msg[0] + 256 * msg[1];
 	appl = msg[2] + 256 * msg[3];
 	nctr = msg[8] & 127;
@@ -1422,9 +1405,9 @@ void init (
 	void  (* rel) (void *),
 	void  (* dwn) (void)
 ) {
-	assert (reg);
-	assert (rel);
-	assert (dwn);
+	assert (reg != NULL);
+	assert (rel != NULL);
+	assert (dwn != NULL);
 
 	fdslusb_capi_ctx->length   = len;
 	fdslusb_capi_ctx->reg_func = reg;
@@ -1488,33 +1471,28 @@ static unsigned	controller_add (void) {
 	
 /*---------------------------------------------------------------------------*\
 \*-M-------------------------------------------------------------------------*/
-static void clear_ubi (dev_context_p pdc) {
+static void clear_ubf (void) {
 	int	ix;
 
-	UNUSED_ARG (pdc);
 	for (ix = 0; ix <= MAX_URB_COUNT; ix++) {
-		ubi[ix].buf = NULL;
-		ubi[ix].urb = NULL;
-		ubi[ix].idx = MAX_URB_COUNT;
+		UBF[ix].buf  = NULL;
+		UBF[ix].urb  = NULL;
+		UBF[ix].flag = URB_UNLINKED;
+		free_ubf[ix] = MAX_URB_COUNT;
+		todo_ubf[ix] = MAX_URB_COUNT;
 	}
-	ubi_n    = -1;
-#if !defined (NDEBUG)
-	ubi_nmin =
-#endif
-	ubi_free = 
-	ubi_in   = 
-	ubi_out  = MAX_URB_COUNT;
-} /* clear_ubi */
+	atomic_set (&free_n, -1);
+} /* clear_ubf */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-int init_urbs (dev_context_p pdc) {
+int init_urbs (void) {
 	int	ix = MAX_URB_COUNT - 1;
 
-	assert (!atomic_read (&pdc->is_open));
 	do {	
-		ubi[ix].idx = 1 + ix;
-		if (NULL == (ubi[ix].buf = hmalloc (MAX_TRANSFER_SIZE))) {
+		free_ubf[ix] = ix;
+		UBF[ix].flag = URB_DONE;
+		if (NULL == (UBF[ix].buf = hmalloc (MAX_TRANSFER_SIZE))) {
 			log (
 				"Could not alloc all %d buffers.\n", 
 				MAX_URB_COUNT
@@ -1522,176 +1500,154 @@ int init_urbs (dev_context_p pdc) {
 			++ix;
 	free_buffers:
 			while (ix < MAX_URB_COUNT) {
-				hfree (ubi[ix++].buf);
+				hfree (UBF[ix++].buf);
 			}
-			clear_ubi (pdc);
+			clear_ubf ();
 			return 0;
 		}
 	} while (--ix >= 0);
 	
 	do {
 		++ix;
-		if (NULL == (ubi[ix].urb = usb_alloc_urb (0))) {
+		if (NULL == (UBF[ix].urb = usb_alloc_urb (0))) {
 			log (
 				"Could not alloc all %d URBs.\n", 
 				MAX_URB_COUNT
 			);
 			--ix;
 			while (ix >= 0) {
-				usb_free_urb (ubi[ix--].urb);
+				usb_free_urb (UBF[ix--].urb);
 			}
 			ix = 0;
 			goto free_buffers;
 		}
-		ubi[ix].urb->context = (void *) ix;
+		UBF[ix].urb->context = (void *) ix;
 	} while (ix < MAX_URB_COUNT);
 	assert (ix == MAX_URB_COUNT);
-	ubi[ix].urb = NULL;
-	ubi[ix].buf = NULL;
-	ubi[ix].idx = MAX_URB_COUNT;
-	
+	UBF[ix].urb = NULL;
+	UBF[ix].buf = NULL;
 	log ("Allocated %d URBs/buffers.\n", ix);
-	ubi_free = 0;
-#if !defined (NDEBUG)
-	ubi_nmin =
-#endif
-	ubi_n    =
-	ubi_in   = 
-	ubi_out  = MAX_URB_COUNT;
+	
+	free_in  = 0;
+	free_out = 0;
+	todo_in  = 0;
+	todo_out = 0;
+	atomic_set (&free_n, MAX_URB_COUNT);
+	atomic_set (&todo_n, 0);
 	return 1;
 } /* init_urbs */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-void free_urbs (dev_context_p pdc) {
+void free_urbs (void) {
 	int	ix;
 	
-	assert (!atomic_read (&pdc->is_open));
-	info (-1 != ubi_n);
-	if (-1 == ubi_n) 
+	info (atomic_read (&free_n) > -1);
+	if (-1 == atomic_read (&free_n)) {
 		return;
+	}
 	for (ix = 0; ix < MAX_URB_COUNT; ix++) {
-		assert (ubi[ix].buf != NULL);
-		assert (ubi[ix].urb != NULL);
-		hfree (ubi[ix].buf);		
-		usb_free_urb (ubi[ix].urb);	
+		assert (UBF[ix].buf != NULL);
+		assert (UBF[ix].urb != NULL);
+		hfree (UBF[ix].buf);		
+		usb_free_urb (UBF[ix].urb);	
 	}
 	log ("Freed %d URBs/buffers.\n", ix);
-	clear_ubi (pdc);
+	clear_ubf ();
 } /* free_urbs */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-void unlink_urbs (dev_context_p pdc) {
+void unlink_urbs (void) {
 	int	ix;
 
-	UNUSED_ARG (pdc);
-	info (-1 != ubi_n);
-	if (-1 == ubi_n) 
+	info (-1 != atomic_read (&free_n));
+	if (-1 == atomic_read (&free_n)) {
 		return;
+	}
 	for (ix = 0; ix < MAX_URB_COUNT; ix++) {
-		if (ubi[ix].idx == URB_SUBMITTED) {
-			usb_unlink_urb (ubi[ix].urb);
-			ubi[ix].idx = URB_UNLINKED;
+		if (UBF[ix].flag == URB_SUBMITTED) {
+			usb_unlink_urb (UBF[ix].urb);
+			UBF[ix].flag = URB_UNLINKED;
 		}
 	}
 } /* unlink_urbs */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-static int claim_urb (
-	dev_context_p	pdc, 
-	struct urb **	ppurb, 
-	char **		ppbuf
-) {
-	int		ix;
-	unsigned long	flags;
+static int claim_urb (struct urb ** ppurb, char ** ppbuf) {
+	int	ix;
 	
-	UNUSED_ARG (pdc);
+	if (!atomic_read (&free_n)) {
+		return 0;
+	}
 	assert (ppurb != NULL);
 	assert (ppbuf != NULL);
-	spin_lock_irqsave (&ullock, flags);
-	ix = ubi_free;
+	atomic_dec (&free_n);
+	ix = free_ubf[free_out];
 	assert (ix >= 0);
-	assert (ix <= MAX_URB_COUNT);
-	ubi_free = ubi[ix].idx;
-	if (ix != MAX_URB_COUNT) {
-		ubi[ix].idx = URB_SUBMITTED;	/* in spe */
-		--ubi_n;
-#if !defined (NDEBUG)
-		if (ubi_n < ubi_nmin) {
-			ubi_nmin = ubi_n;
-			log ("New minimum: %d URBs left.\n", ubi_n);
-		}
-#endif
-	}
-	spin_unlock_irqrestore (&ullock, flags);
-	*ppurb = ubi[ix].urb;
-	*ppbuf = ubi[ix].buf;
-	return (ix != MAX_URB_COUNT);
+	assert (ix < MAX_URB_COUNT);
+	assert (UBF[ix].flag == URB_DONE);
+	free_out = (free_out + 1) % MAX_URB_COUNT;
+	UBF[ix].flag = URB_SUBMITTED;	/* in spe */
+	*ppurb = UBF[ix].urb;
+	*ppbuf = UBF[ix].buf;
+	return 1;
 } /* claim_urb */
 	
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-static void release_urb (dev_context_p pdc, struct urb * purb) {
+static void release_urb (struct urb * purb) {
 	int	ix;
 	
-	UNUSED_ARG (pdc);
 	assert (purb != NULL);
-	assert (in_interrupt ());
 	ix = (int) purb->context;
 	assert (ix >= 0);
 	assert (ix < MAX_URB_COUNT);
-	ubi[ix].idx = ubi_free;
-	ubi_free = ix;
-	++ubi_n;
+	UBF[ix].flag = URB_DONE;
+	free_ubf[free_in] = ix;
+	free_in = (free_in + 1) % MAX_URB_COUNT;
+	assert (atomic_read (&free_n) != MAX_URB_COUNT);
+	atomic_inc (&free_n);
 } /* release_urb */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-static void enlist_urb (dev_context_p pdc, struct urb * purb) {
+static void enlist_urb (struct urb * purb) {
 	int	ix;
 
-	assert (pdc != NULL);
 	assert (purb != NULL);
-	assert (in_interrupt ());
 	ix = (int) purb->context;
 	assert (ix >= 0);
 	assert (ix < MAX_URB_COUNT);
-	ubi[ix].idx = MAX_URB_COUNT;
-	if (ubi_in != MAX_URB_COUNT) {
-		ubi[ubi_in].idx = ix;
-	} else {
-		ubi_out = ix;
-	}
-	ubi_in = ix;
-	atomic_inc (&pdc->rx_count);
+	assert (UBF[ix].flag == URB_SUBMITTED);
+	UBF[ix].flag = URB_COMPLETED;
+	todo_ubf[todo_in] = ix;
+	todo_in = (todo_in + 1) % MAX_URB_COUNT;
+	assert (atomic_read (&todo_n) != MAX_URB_COUNT);
+	atomic_inc (&todo_n);
 } /* enlist_urb */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-static int unlist_urb (
-	dev_context_p	pdc,
-	struct urb **	ppurb,
-	char **		ppbuf
-) {
-	int		ix;
+static int unlist_urb (struct urb ** ppurb, char ** ppbuf) {
+	int	ix;
 
-	assert (pdc != NULL);
+	if (!atomic_read (&todo_n)) {
+		return 0;
+	}
+	atomic_dec (&todo_n);
 	assert (ppurb != NULL);
 	assert (ppbuf != NULL);
-	assert (in_interrupt ());
-	ix = ubi_out;
+	ix = todo_ubf[todo_out];
 	assert (ix >= 0);
-	assert (ix <= MAX_URB_COUNT);
-	ubi_out = ubi[ix].idx;
-	if (ubi_out == MAX_URB_COUNT) {
-		ubi_in = MAX_URB_COUNT;
-	}
-	atomic_dec (&pdc->rx_count);
-	*ppurb = ubi[ix].urb;
-	*ppbuf = ubi[ix].buf;
-	return (ix != MAX_URB_COUNT);
+	assert (ix < MAX_URB_COUNT);
+	assert (UBF[ix].flag == URB_COMPLETED);
+	UBF[ix].flag = URB_PROCESSING;
+	todo_out = (todo_out + 1) % MAX_URB_COUNT;
+	*ppurb = UBF[ix].urb;
+	*ppbuf = UBF[ix].buf;
+	return 1;
 } /* unlist_urb */
 
 /*---------------------------------------------------------------------------*\
@@ -1699,65 +1655,38 @@ static int unlist_urb (
 static void issue_hard_error (dev_context_p pdc, const char * msg, int arg) {
 
 	lprintf (KERN_INFO, msg, arg);
-	enter_critical ();
 	if (!hard_error_issued) {
 		hard_error_issued = TRUE;
 		(*pdc->ev_handler) (USB_EVENT_HARD_ERROR);
 	} else {
 		lprintf (KERN_INFO, "Extra events ignored.\n");
 	}
-	leave_critical ();
 } /* issue_hard_error */
 
 /*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
-static void tx_dpc (void * data) {
-	static void	usb_write (dev_context_p, unsigned long);
-	dev_context_p	pdc;
-	struct urb *	purb;
-	
-	assert (in_interrupt ());
-	assert (data != NULL);
-	pdc = (dev_context_p) data;
-	purb = pdc->tx_urb;
-	assert (atomic_read (&pdc->tx_pending));
-	if (!atomic_read (&pdc->is_open)) {
-		atomic_set (&pdc->tx_pending, 0);
-		return;
-	}
-	if (0 != purb->status) {
-		atomic_set (&pdc->tx_pending, 0);
-		issue_hard_error (pdc, "Tx URB status: %d\n", purb->status);
-		return;
-	}
-	if ((purb->actual_length > 0) && ((purb->actual_length % 64) == 0)) {
-		log (
-			"Inserting zero-length data block! (Wrote %d bytes)\n",
-			purb->actual_length
-		);
-		usb_write (pdc, 0);
-	} else {
-		atomic_set (&pdc->tx_pending, 0);
-		usb_TxStart ();	
-		kick_scheduler ();
-	}
-} /* tx_dpc */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
+\*---------------------------------------------------------------------------*/
 static void usb_write_completion (struct urb * purb) {
 
 	UNUSED_ARG (purb);
-	assert (in_interrupt ());
-	kick_queue (&tq_tx_dpc);
+	tasklet_schedule (&tx_tasklet);
 } /* usb_write_completion */
 
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static void usb_read_completion (struct urb * purb) {
+
+	assert (purb != NULL);
+	atomic_dec (&fdslusb_capi_ctx->rx_pending);
+	enlist_urb (purb);
+	tasklet_schedule (&rx_tasklet);
+} /* usb_read_completion */
+
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/ 
 static void usb_write (dev_context_p pdc, unsigned long len) {
 	int	res;
 
-	assert (pdc);
+	assert (pdc != NULL);
 	if (!atomic_read (&pdc->is_open)) {
 		log ("Attempt to write on closed device.");
 		return;
@@ -1778,48 +1707,6 @@ static void usb_write (dev_context_p pdc, unsigned long len) {
 	}
 } /* usb_write */
 
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
-static void rx_dpc (void * data) {
-	dev_context_p	pdc;
-	struct urb *	purb;
-	char *		pbuf;
-
-	assert (in_interrupt ());
-	assert (data != NULL);
-	pdc = (dev_context_p) data;
-	if (!atomic_read (&pdc->is_open)) {
-		return;
-	}
-	if (!unlist_urb (pdc, &purb, &pbuf)) {
-		log ("Completion w/o URB!\n");
-		return;
-	}
-	if (0 != purb->status) {
-		issue_hard_error (pdc, "Rx URB status: %d\n", purb->status);
-		return;
-	}
-	usb_RxStart ();
-	if (0 != purb->actual_length) {
-		enter_critical ();
-		(*pdc->rx_avail) (pbuf, purb->actual_length);
-		leave_critical ();
-		kick_scheduler ();
-	}
-	release_urb (pdc, purb);
-} /* rx_dpc */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
-static void usb_read_completion (struct urb * purb) {
-
-	assert (in_interrupt ());
-	assert (purb != NULL);
-	atomic_dec (&fdslusb_capi_ctx->rx_pending);
-	enlist_urb (fdslusb_capi_ctx, purb);
-	kick_queue (&tq_rx_dpc);
-} /* usb_read_completion */
-
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 static void usb_read (dev_context_p pdc) {
@@ -1827,12 +1714,12 @@ static void usb_read (dev_context_p pdc) {
 	char *		pbuf;
 	int		res;
 
-	assert (pdc);
+	assert (pdc != NULL);
 	if (!atomic_read (&pdc->is_open)) {
 		log ("Attempt to read from closed device.\n");
 		return;
 	}
-	if (!claim_urb (pdc, &purb, &pbuf)) {
+	if (!claim_urb (&purb, &pbuf)) {
 		log ("No URB/buffer available.\n");
 		return;
 	}
@@ -1853,6 +1740,99 @@ static void usb_read (dev_context_p pdc) {
 	}
 } /* usb_read */
 
+/*---------------------------------------------------------------------------*\
+\*-T-------------------------------------------------------------------------*/ 
+static void tx_handler (dev_context_p pdc) {
+	static void	usb_write (dev_context_p, unsigned long);
+	struct urb *	purb;
+	
+	assert (pdc != NULL);
+	purb = pdc->tx_urb;
+	assert (atomic_read (&pdc->tx_pending));
+	atomic_set (&tx_flag, 0);
+	if (!atomic_read (&pdc->is_open)) {
+		atomic_set (&pdc->tx_pending, 0);
+		log ("TX: Device is closed!\n");
+		return;
+	}
+	if (0 != purb->status) {
+		atomic_set (&pdc->tx_pending, 0);
+		issue_hard_error (pdc, "Tx URB status: %d\n", purb->status);
+		return;
+	}
+	if ((purb->actual_length > 0) && ((purb->actual_length % 64) == 0)) {
+		log (
+			"Inserting zero-length data block! (Wrote %d bytes)\n",
+			purb->actual_length
+		);
+		usb_write (pdc, 0);
+	} else {
+		atomic_set (&pdc->tx_pending, 0);
+		usb_TxStart ();	
+		kick_scheduler ();
+	}
+} /* tx_handler */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/ 
+static void tx_task (unsigned long data) {
+
+	UNUSED_ARG (data);
+	if (in_critical ()) {
+		atomic_set (&tx_flag, 1);
+	} else if (spin_trylock (&stack_lock)) {
+		tx_handler (fdslusb_capi_ctx);
+		spin_unlock (&stack_lock);
+	} else {
+		atomic_set (&tx_flag, 1);
+	}
+} /* tx_task */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/ 
+static void rx_handler (dev_context_p pdc) {
+	struct urb *	purb;
+	char *		pbuf;
+
+	assert (pdc != NULL);
+	atomic_set (&rx_flag, 0);
+	if (!atomic_read (&pdc->is_open)) {
+		log ("RX: Device is closed!\n");
+		return;
+	}
+	if (!unlist_urb (&purb, &pbuf)) {
+		log ("Completion w/o URB!\n");
+		return;
+	}
+	if (0 != purb->status) {
+		issue_hard_error (pdc, "Rx URB status: %d\n", purb->status);
+		return;
+	}
+	usb_RxStart ();
+	if (0 != purb->actual_length) {
+		enter_critical ();
+		(*pdc->rx_avail) (pbuf, purb->actual_length);
+		leave_critical ();
+		kick_scheduler ();
+	}
+	release_urb (purb);
+} /* rx_handler */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/ 
+static void rx_task (unsigned long data) {
+
+	UNUSED_ARG (data);
+	if (in_critical ()) {
+		atomic_set (&rx_flag, 1);
+	} else if (spin_trylock (&stack_lock)) {
+		rx_handler (fdslusb_capi_ctx);
+		spin_unlock (&stack_lock);
+	} else {
+		atomic_set (&rx_flag, 1);
+	}
+} /* rxi_task */
+
 /*---------------------------------------------------------------------------*\
 \*-S-------------------------------------------------------------------------*/
 static void scheduler_control (unsigned f) {
@@ -1872,6 +1852,62 @@ static void scheduler_control (unsigned f) {
 	}
 } /* scheduler_control */
 
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static int scheduler (void * arg) {
+
+	UNUSED_ARG (arg);
+	SNPRINTF (current->comm, 16, "%s_thread", TARGET);	/* See: sched.h */
+	log ("Starting scheduler thread '%s'...\n", current->comm);
+	while (atomic_read (&thread_run_flag)) {
+		if (!atomic_read (&thread_capi_run_flag)) {
+			interruptible_sleep_on (&capi_wait);
+		} else {
+			if (atomic_read (&scheduler_enabled)) {
+				interruptible_sleep_on_timeout (&wait, TEN_MSECS);
+			} else {
+				interruptible_sleep_on (&wait);
+			}
+		}
+		if (!atomic_read (&thread_run_flag)) {
+			info (atomic_read (&thread_run_flag));
+			break;
+		}
+		if (!atomic_read (&thread_capi_run_flag)) {
+			info (atomic_read (&thread_capi_run_flag));
+			continue;
+		}
+		if (!atomic_read (&scheduler_enabled)) {
+			info (atomic_read (&scheduler_enabled));
+			continue;
+		}
+		if (fdslusb_capi_lib == NULL) {
+			log ("Library lost! Hot unplug?\n");
+			break;
+		}
+		if (spin_trylock (&stack_lock)) {
+			os_timer_poll ();
+			if ((*fdslusb_capi_lib->cm_schedule) ()) {
+				scheduler_control (TRUE);
+			}
+			spin_unlock (&stack_lock);
+		} else {
+			kick_scheduler ();
+		}
+	}
+	log ("Scheduler thread stopped.\n");
+	up (&hotplug);
+	return 0;
+} /* scheduler */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static void kick_scheduler (void) {
+
+	atomic_set (&scheduler_enabled, 1);
+	wake_up_interruptible (&wait);
+} /* kick_scheduler */
+
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 int make_thread (void) {
@@ -1934,67 +1970,6 @@ static void disable_thread (void) {
 	}
 } /* disable_thread */
 
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/
-static void dpc (void * data) {
-
-	UNUSED_ARG (data);
-	if (!in_critical ()) {
-		run_task_queue (&tq_private);
-	}
-} /* dpc */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/
-static int scheduler (void * arg) {
-
-	UNUSED_ARG (arg);
-	SNPRINTF (current->comm, 16, "%s_thread", TARGET);	/* See: sched.h */
-	log ("Starting scheduler thread '%s'...\n", current->comm);
-	while (atomic_read (&thread_run_flag)) {
-		if (!atomic_read (&thread_capi_run_flag)) {
-			interruptible_sleep_on (&capi_wait);
-		} else {
-			if (atomic_read (&scheduler_enabled)) {
-				interruptible_sleep_on_timeout (&wait, TEN_MSECS);
-			} else {
-				interruptible_sleep_on (&wait);
-			}
-		}
-		if (!atomic_read (&thread_run_flag)) {
-			info (atomic_read (&thread_run_flag));
-			break;
-		}
-		if (!atomic_read (&thread_capi_run_flag)) {
-			info (atomic_read (&thread_capi_run_flag));
-			continue;
-		}
-		if (!atomic_read (&scheduler_enabled)) {
-			info (atomic_read (&scheduler_enabled));
-			continue;
-		}
-		if (fdslusb_capi_lib == NULL) {
-			log ("Library lost! Hot unplug?\n");
-			break;
-		}
-		os_timer_poll ();
-		if ((*fdslusb_capi_lib->cm_schedule) ()) {
-			scheduler_control (TRUE);
-		}
-	}
-	log ("Scheduler thread stopped.\n");
-	up (&hotplug);
-	return 0;
-} /* scheduler */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/
-static void kick_scheduler (void) {
-
-	atomic_set (&scheduler_enabled, 1);
-	wake_up_interruptible (&wait);
-} /* kick_scheduler */
-
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 static void stop_base (void) {
@@ -2006,6 +1981,8 @@ static void stop_base (void) {
 	atomic_set (&pdc->is_open, 0);
 	atomic_set (&pdc->rx_pending, 0);
 	atomic_set (&pdc->tx_pending, 0);
+	atomic_set (&rx_flag, 0);
+	atomic_set (&tx_flag, 0);
 	log ("USB layer stopped.\n");
 } /* stop_base */
 		
@@ -2024,7 +2001,6 @@ void usb_Open (
 	atomic_set (&pdc->is_open, 1);
 	atomic_set (&pdc->tx_pending, 0);
 	atomic_set (&pdc->rx_pending, 0);
-	atomic_set (&pdc->rx_count, 0);
 	pdc->max_buf_size = max_buffer_size;
 	pdc->tx_get_next  = get_next_tx_buffer;
 	pdc->rx_avail     = new_rx_buffer_avail;
@@ -2040,7 +2016,7 @@ void usb_Open (
 void usb_Close (complete_func_p close_complete, unsigned ref_data) {
 
 	stop_base ();
-	assert (close_complete);
+	assert (close_complete != NULL);
 	(*close_complete) (0, ref_data);
 	if (hard_error_issued) {
 		start_closing_task (NULL, NULL);
@@ -2053,9 +2029,9 @@ static void usb_RxStart (void) {
 	dev_context_p	pdc;
 
 	pdc = fdslusb_capi_ctx;
-	assert (pdc);
-	if (atomic_read (&pdc->rx_count) > 0) {
-	 	kick_queue (&tq_rx_dpc);
+	assert (pdc != NULL);
+	if (atomic_read (&todo_n) > 0) {
+		tasklet_schedule (&rx_tasklet);
 	} else {
 	 	usb_read (pdc);
 	}
@@ -2067,7 +2043,7 @@ void usb_TxStart (void) {
 	dev_context_p	pdc = fdslusb_capi_ctx;
 	unsigned long	len;
 
-	assert (pdc);
+	assert (pdc != NULL);
 	if (atomic_xchg (&pdc->tx_pending, 1)) {
 		return;
 	}
